/*
 * Copyright (c) 1995-1998 Greg Ungerer (gerg@stallion.oz.au).
 * Copyright (c) 2001 Murray Stokely
 * Copyright (c) 2001 Wind River Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Greg Ungerer.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: stallion.c,v 1.12 2001/09/23 09:49:20 murray Exp $
 */

/*
 * This is a device driver for PCI Easy-IO devices from Stallion technologies
 */

/*****************************************************************************/

#include <sys/param.h>  /* defines used in kernel.h */
#include <sys/proc.h>   /* For suser */
#include <sys/ucred.h>  /* For suser */
#include <sys/module.h>
#include <sys/systm.h>  /* uprintf */ 
#include <sys/errno.h>
#include <sys/kernel.h> /* types used in module initialization */
#include <sys/conf.h>  /* cdevsw, d_open_t, types */
#include <sys/malloc.h>
#include <sys/fcntl.h>		/* So we can get open modes */
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>	/* structs, prototypes for pci bus stuff */
#include <sys/rman.h>
#include <sys/tty.h>
#define TTYDEFCHARS
#include <sys/ttydefaults.h>

#include <pci/pcivar.h> /* For get_pci macros! */
#include <pci/pcireg.h> /* For PCIR_ defines */

#include <dev/stl/comstats.h>
#include <dev/stl/stallion.h>

/* Function prototypes */
d_open_t      stlopen;
d_close_t     stlclose;
d_read_t      stl_read;
d_write_t     stl_write;
d_ioctl_t     stl_ioctl;

int stl_initeio(struct stl_softc *sc);
int stl_initports(struct stl_softc *sc, stlpanel_t *panelp, int strtport);
void stl_eiointr(void *xsc);
static int  stl_rawopen(stlport_t *portp);
static void stl_start(struct tty *tp);
static int  stl_param(struct tty *tp, struct termios *tiosp);
static void stl_flowcontrol(stlport_t *portp, int hw, int sw);
static int  stl_rawclose(stlport_t *portp);
static void stl_flush(stlport_t *portp, int flag);
static void stl_dtrwakeup(void *arg);
static void	stl_stop(struct tty *tp, int);

/* Character device entry points */

static struct cdevsw stl_cdevsw = {
	stlopen,
	stlclose,
/*	stl_read, */
	ttyread,
	ttywrite,
/*	stl_write, */
	stl_ioctl,
	nopoll,
	nommap,
	nostrategy,
	"stl",
	72,                   /* major used by old stl */
	nodump,
	nopsize,
	D_TTY,
/* XXX this is not tested, -1 works but causes a compiler warning. maybe this works? */
	-1
/*	NULL, */
/*    -1 */
};

/* vars */
devclass_t stl_devclass;
static int		stl_doingtimeout = 0;

/*
 *	Define some macros to make calling these functions nice and clean.
 */
#define	stl_panelinit		(* ((uart_t *) panelp->uartp)->panelinit)
#define	stl_portinit		(* ((uart_t *) portp->uartp)->portinit)
#define	stl_setport		(* ((uart_t *) portp->uartp)->setport)
#define	stl_getsignals		(* ((uart_t *) portp->uartp)->getsignals)
#define	stl_setsignals		(* ((uart_t *) portp->uartp)->setsignals)
#define	stl_enablerxtx		(* ((uart_t *) portp->uartp)->enablerxtx)
#define	stl_startrxtx		(* ((uart_t *) portp->uartp)->startrxtx)
#define	stl_disableintrs	(* ((uart_t *) portp->uartp)->disableintrs)
#define	stl_sendbreak		(* ((uart_t *) portp->uartp)->sendbreak)
#define	stl_sendflow		(* ((uart_t *) portp->uartp)->sendflow)
#define	stl_uartflush		(* ((uart_t *) portp->uartp)->flush)
#define	stl_datastate		(* ((uart_t *) portp->uartp)->datastate)

/*****************************************************************************/

/*
 *	Interrupt service routine for EasyIO boards.
 */

void
stl_eiointr(void *xsc) {
	uprintf("stl_eiointr!\n");
/*	stlpanel_t	*panelp;
	int		iobase; */

	return;
/*
	panelp = (stlpanel_t *) sc->panels[0];
	iobase = panelp->iobase;
	while (inb(brdp->iostatus) & EIO_INTRPEND)
		(* panelp->isr)(panelp, iobase);
*/
}

int
stl_makedevs(struct stl_softc *sc) {
	int port = 0;

	/* XXX What does this device even do? */
	/* XXX bogus minor number */
	sc->iomemdev = make_dev(&stl_cdevsw, (99999), UID_ROOT, GID_WHEEL, 
	    0600, "staliomem%d", sc->unit);

	/* Create serial devices for each port */
	/* XXX this is funky only works for 1 board */
	for (port = 0; port < sc->nrports; port++) {
		uprintf("Creating devices for port : %d\n", port);
		/* ttyE device */
		sc->ttyE[port] = make_dev(&stl_cdevsw, STL_MINOR(sc->unit, 
		    port, 0), UID_ROOT, GID_WHEEL, 0600, "ttyE%d%d", 
		    sc->unit, port);
		/* ttyiE device */
		sc->ttyiE[port] = make_dev(&stl_cdevsw, STL_MINOR(sc->unit,
		    port, 1), UID_ROOT, GID_WHEEL, 0600, "ttyiE%d%d", 
		    sc->unit, port);
		/* ttylE device */
		sc->ttylE[port] = make_dev(&stl_cdevsw, STL_MINOR(sc->unit, 
		    port, 2), UID_ROOT, GID_WHEEL, 0600, "ttylE%d%d", 
		    sc->unit, port);
		/* cue device */
		sc->cue[port] = make_dev(&stl_cdevsw, STL_MINOR(sc->unit, 
		    port, 3), UID_ROOT, GID_WHEEL, 0600, "cue%d%d", 
		    sc->unit, port);
		/* cuie device */
		sc->cuie[port] = make_dev(&stl_cdevsw, STL_MINOR(sc->unit,
		    port, 4), UID_ROOT, GID_WHEEL, 0600, "cuie%d%d", 
		    sc->unit, port);
		/* cule device */
		sc->cule[port] = make_dev(&stl_cdevsw, STL_MINOR(sc->unit,
		    port, 5), UID_ROOT, GID_WHEEL, 0600, "cule%d%d", 
		    sc->unit, port);
	}
	return 0;
}

/***************************************************************/
/* KSE Milestone 2 makes us think about threads, not processes */
int
#if __FreeBSD_version >= 500023
stlopen(dev_t dev, int flag, int mode, struct thread *td)
#else
stlopen(dev_t dev, int flag, int mode, struct proc *p)
#endif
{
        struct stl_softc *sc;
	struct tty	*tp;
	stlport_t	*portp;
	int		error, callout, x;
	int portnum, devnum;
	device_t device = devclass_get_device(stl_devclass, 
	    MINORTOUNIT(minor(dev)));

	sc = (struct stl_softc *)device_get_softc(device);
	/* Which port does this open refer to? */
	portnum = MINORTOPORT(sc->unit, minor(dev));
	devnum  = MINORTODEV(sc->unit, minor(dev));
	uprintf("This is an open of the %d device %d port\n", devnum, 
	    portnum);
	portp = sc->ports[portnum];

	if (portp == (stlport_t *) NULL)
		return(ENXIO);

/* XXX Determine if this is the ctrl device? */
/*
	if (dev & STL_CTRLDEV)
		return(0);
*/

	tp = &portp->tty;
	/* Is this a callout device? */
	if ((devnum == 3) || (devnum == 4) || (devnum == 5))
		callout = 1;
	else
		callout = 0;
	error = 0;
	uprintf("spltty\n");
	x = spltty();

stlopen_restart:
	uprintf("stlopen_restart\n");
/*
 *	Wait here for the DTR drop timeout period to expire.
 */

	while (portp->state & ASY_DTRWAIT) {
		error = tsleep(&portp->dtrwait, (TTIPRI | PCATCH),
		    "stldtr", 0);
		if (error)
			goto stlopen_end;
	}

/*
 *	We have a valid device, so now we check if it is already open.
 *	If not then initialize the port hardware and set up the tty
 *	struct as required.
 */
	uprintf("Check t_state if its open\n");
	if ((tp->t_state & TS_ISOPEN) == 0) {
		tp->t_oproc = stl_start;
		tp->t_stop  = stl_stop;
		tp->t_param = stl_param;
		tp->t_dev = dev;
		tp->t_termios = callout ? portp->initouttios :
		    portp->initintios;
		stl_rawopen(portp);
		ttsetwater(tp);
		if ((portp->sigs & TIOCM_CD) || callout)
			(*linesw[tp->t_line].l_modem)(tp, 1);
	} else {
		if (callout) {
			if (portp->callout == 0) {
				error = EBUSY;
				goto stlopen_end;
			}
		} else {
			if (portp->callout != 0) {
				if (flag & O_NONBLOCK) {
					error = EBUSY;
					goto stlopen_end;
				}
				error = tsleep(&portp->callout,
				    (TTIPRI | PCATCH), "stlcall", 0);
				if (error)
					goto stlopen_end;
				goto stlopen_restart;
			}
		}
/* whats wrong here? */
		if ((tp->t_state & TS_XCLUDE) && (SUSER)) {
			error = EBUSY;
			goto stlopen_end;
		}
	}
	
/*
 *	If this port is not the callout device and we do not have carrier
 *	then we need to sleep, waiting for it to be asserted.
 */
	uprintf("If this is not the callout device and we don't have carrier,"
	        " then we need to sleep\n");
	if (((tp->t_state & TS_CARR_ON) == 0) && !callout &&
	    ((tp->t_cflag & CLOCAL) == 0) && ((flag & O_NONBLOCK) == 0)) {
	        uprintf("Yes that is the case, entering tsleep()!\n");
		portp->waitopens++;
		error = tsleep(TSA_CARR_ON(tp), (TTIPRI | PCATCH), "stldcd", 
		    0);
	        uprintf("Returned from tsleep: %d\n", error);
		portp->waitopens--;
		if (error)
			goto stlopen_end;
		goto stlopen_restart;
	}

/*
 *	Open the line discipline.
 */

	uprintf("open line discipline\n");
	error = (*linesw[tp->t_line].l_open)(dev, tp);
	stl_ttyoptim(portp, &tp->t_termios);
	if ((tp->t_state & TS_ISOPEN) && callout)
		portp->callout = 1;

/*
 *	If for any reason we get to here and the port is not actually
 *	open then close of the physical hardware - no point leaving it
 *	active when the open failed...
 */

stlopen_end:
	uprintf("About to call splx!\n");
	splx(x);
	uprintf("splx returned!\n");
	if (((tp->t_state & TS_ISOPEN) == 0) && (portp->waitopens == 0)) {
		uprintf("port is not actually open?  Calling rawclose!\n");
		stl_rawclose(portp);
	}
	uprintf("stlopen finished, about to return : %d\n", error);
	return(error);
}


/*****************************************************************************/
/* KSE Milestone 2 makes us think about threads, not processes */
int
#if __FreeBSD_version >= 500023
stlclose(dev_t kdev, int flag, int mode, struct thread *td)
#else
stlclose(dev_t kdev, int flag, int mode, struct proc *p)
#endif
{
	struct stl_softc *sc;
	struct tty	*tp;
	stlport_t	*portp;
	int		x;
	int devnum, portnum;
	device_t dev = devclass_get_device(stl_devclass, 
	    MINORTOUNIT( minor(kdev) ));
	sc = (struct stl_softc *)device_get_softc(dev);
	uprintf("stlclose(dev=%p, flag=%x, mode=%x)\n", &dev, flag, mode);
	portnum = MINORTOPORT(sc->unit, minor(kdev));
	devnum  = MINORTODEV(sc->unit, minor(kdev));
	uprintf("This is a close of the %d device %d port\n", devnum, portnum);
	portp = sc->ports[portnum];

/* XXX We should fix this.
	if (dev & STL_MEMDEV)
		return(0);
	if (dev & STL_CTRLDEV)
		return(0);
*/

	if (portp == (stlport_t *) NULL)
		return(ENXIO);
	tp = &portp->tty;

	uprintf("About to call stltty\n");
	x = spltty();
	uprintf("About to call function pointer of some kind!\n");
/*
  this succeeeds above :
	error = (*linesw[tp->t_line].l_open)(dev, tp);
*/
	uprintf("Function Ptr : %p\n", *linesw[tp->t_line].l_close);
	(*linesw[tp->t_line].l_close)(tp, -99);
/* flag); */
	uprintf("Returned from function ptr!\n");
	return 0;

	uprintf("About to call ttyoptim!\n");
	stl_ttyoptim(portp, &tp->t_termios);
	uprintf("About to call rawclose!\n");
	stl_rawclose(portp);
	uprintf("About to call ttyclose!\n");
	ttyclose(tp);
	uprintf("About to call splx!\n");	
	splx(x);
	uprintf("Returning 0!\n");
	return(0);
}

int
stl_read(dev_t kdev, struct uio *uio, int ioflag)
{
	struct stl_softc *sc;
	device_t dev = devclass_get_device(stl_devclass, minor(kdev) / 16);
	int err = 0;
/*    uint16_t i; */

	sc = (struct stl_softc *)device_get_softc(dev);

/*    switch (minor(kdev) % 16) { */

	return err;
}

int
stl_write(dev_t kdev, struct uio *uio, int ioflag)
{
	struct stl_softc *sc;
	device_t dev = devclass_get_device(stl_devclass, minor(kdev) / 16);
	int err = 0;

	sc = (struct stl_softc *)device_get_softc(dev);

	return err;
}


int
#if __FreeBSD_version >= 500023
stl_ioctl(dev_t kdev, u_long cmd, caddr_t data, int fflag, struct thread *td)
#else
stl_ioctl(dev_t kdev, u_long cmd, caddr_t data, int fflag, struct proc *p)
#endif
{
/*    struct stl_softc *sc = devclass_get_softc(stl_devclass, minor(kdev) / 16);*/
/*     struct daq_device *pcidev = (struct stl_device *)data; */

/*    switch(minor(kdev) % 16) { */
	return 0;
}


/*****************************************************************************/

/*
 *	Initialize and configure the specified board. This firstly probes
 *	for the board, if it is found then the board is initialized and
 *	then all its ports are initialized as well.
 */

int stl_brdinit(struct stl_softc *sc)
{
	stlpanel_t	*panelp;
	int		i, j, k;

	uprintf("stl_brdinit - about to call initeio\n");
	stl_initeio(sc);

	for (i = 0, k = 0; (i < STL_MAXPANELS); i++) {
		panelp = sc->panels[i];
		if (panelp != (stlpanel_t *) NULL) {
			stl_initports(sc, panelp, k);
			for (j = 0; (j < panelp->nrports); j++)
				sc->ports[k++] = panelp->ports[j];
		}
	}

	uprintf("stl%d: unit=%d nrpanels=%d nrports=%d\n",
	       sc->unit, sc->unit, sc->nrpanels, sc->nrports);

	return(0);
}

/*
 * Setup IRQ line, find capabilities for this kind of PCI EasyIO card,
 * malloc space for at least one panel of ports and setup the
 * appropriate ISR function pointer.
 */

int stl_initeio(struct stl_softc *sc)
{
	stlpanel_t	*panelp;
	unsigned int	status = 0;
/* This won't work, because we haven't attached yet! */

	uprintf("stl_initeio\n");

/*	brdp->ioctrl = brdp->ioaddr1 + 1;
	brdp->iostatus = brdp->ioaddr1 + 2;
	brdp->clk = EIO_CLK; */

	sc->irqid = 0x0;
	sc->irqres = bus_alloc_resource(sc->dev, SYS_RES_IRQ, &(sc->irqid),
	    0, ~0, 1, RF_SHAREABLE | RF_ACTIVE);
	
	if (sc->irqres == NULL) {
		uprintf("IRQ allocation failed!\n");
		return ENXIO;
	}


	status = bus_setup_intr(sc->dev, sc->irqres, INTR_TYPE_TTY,
	    stl_eiointr, sc, &(sc->handler));

	if ( status ) {
		uprintf("Couldn't set up irq\n");
		return ENXIO;
	}

	status = BAR0_READ1(sc, 2);

	switch (status & EIO_IDBITMASK) {
	case EIO_8PORTM:
		sc->clk = EIO_CLK8M;
		/* fall thru */
	case EIO_8PORTRS:
	case EIO_8PORTDI:
		sc->nrports = 8;
		break;
	case EIO_4PORTRS:
	    uprintf("EIO4portRS!\n");
		sc->nrports = 4;
		break;
	case EIO_MK3:
		switch (status & EIO_BRDMASK) {
		case ID_BRD4:
			uprintf("EIO_MK3 - ID_BRD4!\n");
			sc->nrports = 4;
			break;
		case ID_BRD8:
			sc->nrports = 8;
			break;
		case ID_BRD16:
			sc->nrports = 16;
			break;
		default:
			return(ENODEV);
		}
		sc->ioctrl++;
		break;
	default:
		return(ENODEV);
	}

	BAR1_WRITE1(sc, 0x4c, 0x41);

	/* Create a panel structure */
	panelp = (stlpanel_t *) malloc(sizeof(stlpanel_t), M_TTYS, M_ZERO);
	if (panelp == (stlpanel_t *) NULL) {
		uprintf("STALLION: failed to allocate memory (size=%d)\n",
		    sizeof(stlpanel_t));
		return(ENOMEM);
	}

	panelp->sc      = sc;
	panelp->panelnr = 0;
	panelp->nrports = sc->nrports;
	/* iobase undefined! */
	panelp->hwid = status;
	if ((status & EIO_IDBITMASK) == EIO_MK3) {
		uprintf("EIO MK3 detected\n");
		panelp->uartp = (void *) &stl_sc26198uart;
		panelp->isr = stl_sc26198intr;
	} else {
		uprintf("ERROR The CD1400 UART is not supported with this version.");
		return(ENXIO);
	}
	sc->panels[0] = panelp;
	sc->nrpanels = 1;
	sc->hwid = status;
	return(0);
}

/*****************************************************************************/

/*
 *	Enable l_rint processing bypass mode if tty modes allow it.
 */

void 
stl_ttyoptim(stlport_t *portp, struct termios *tiosp)
{
	struct tty	*tp;

	tp = &portp->tty;
	if (((tiosp->c_iflag & (ICRNL | IGNCR | IMAXBEL | INLCR | ISTRIP)) == 0) &&
	    (((tiosp->c_iflag & BRKINT) == 0) || (tiosp->c_iflag & IGNBRK)) &&
	    (((tiosp->c_iflag & PARMRK) == 0) ||
		((tiosp->c_iflag & (IGNPAR | IGNBRK)) == (IGNPAR | IGNBRK))) &&
	    ((tiosp->c_lflag & (ECHO | ICANON | IEXTEN | ISIG | PENDIN)) ==0) &&
	    (linesw[tp->t_line].l_rint == ttyinput))
		tp->t_state |= TS_CAN_BYPASS_L_RINT;
	else
		tp->t_state &= ~TS_CAN_BYPASS_L_RINT;

	if (tp->t_line == SLIPDISC)
		portp->hotchar = 0xc0;
	else if (tp->t_line == PPPDISC)
		portp->hotchar = 0x7e;
	else
		portp->hotchar = 0;
}



/*****************************************************************************/

/*
 *	Try and find and initialize all the ports on a panel. We don't care
 *	what sort of board these ports are on - since the port io registers
 *	are almost identical when dealing with ports.
 */

int stl_initports(struct stl_softc *sc, stlpanel_t *panelp, int strtport)
{
	stlport_t	*portp;
	unsigned int	chipmask;
	int		i, j;

	uprintf("In stl_initports!\n");

/*#define	stl_panelinit		(* ((uart_t *) panelp->uartp)->panelinit) */
	chipmask = stl_panelinit(sc, panelp);

/*
 *	All UART's are initialized if found. Now go through and setup
 *	each ports data structures. Also initialize each individual
 *	UART port.
 */
	for (i = 0; (i < panelp->nrports); i++) {
		portp = (stlport_t *) malloc(sizeof(stlport_t), M_TTYS,
			M_NOWAIT);
		if (portp == (stlport_t *) NULL) {
			printf("STALLION: failed to allocate port memory "
			    "(size=%d)\n", sizeof(stlport_t));
			break;
		}
		bzero(portp, sizeof(stlport_t));

		portp->portnr = i;
		portp->sc = sc;
		portp->panelnr = panelp->panelnr;
		portp->uartp = panelp->uartp;
		portp->clk = sc->clk;
		panelp->ports[i] = portp;

		j = STL_TXBUFSIZE + (2 * STL_RXBUFSIZE);
		portp->tx.buf = (char *) malloc(j, M_TTYS, M_NOWAIT);
		if (portp->tx.buf == (char *) NULL) {
			printf("STALLION: failed to allocate buffer memory "
			    "(size=%d)\n", j);
			break;
		}
		portp->tx.endbuf = portp->tx.buf + STL_TXBUFSIZE;
		portp->tx.head = portp->tx.buf;
		portp->tx.tail = portp->tx.buf;
		portp->rx.buf = portp->tx.buf + STL_TXBUFSIZE;
		portp->rx.endbuf = portp->rx.buf + STL_RXBUFSIZE;
		portp->rx.head = portp->rx.buf;
		portp->rx.tail = portp->rx.buf;
		portp->rxstatus.buf = portp->rx.buf + STL_RXBUFSIZE;
		portp->rxstatus.endbuf = portp->rxstatus.buf + STL_RXBUFSIZE;
		portp->rxstatus.head = portp->rxstatus.buf;
		portp->rxstatus.tail = portp->rxstatus.buf;
		bzero(portp->rxstatus.head, STL_RXBUFSIZE);

		portp->initintios.c_ispeed = STL_DEFSPEED;
		portp->initintios.c_ospeed = STL_DEFSPEED;
		portp->initintios.c_cflag = STL_DEFCFLAG;
		portp->initintios.c_iflag = 0;
		portp->initintios.c_oflag = 0;
		portp->initintios.c_lflag = 0;
		bcopy(&ttydefchars[0], &portp->initintios.c_cc[0],
		    sizeof(portp->initintios.c_cc));
		portp->initouttios = portp->initintios;
		portp->dtrwait = 3 * hz;

/*#define	stl_portinit		(* ((uart_t *) portp->uartp)->portinit)*/
		stl_portinit(sc, panelp, portp);



/*
 * XXXX THIS IS WHERE WE SHOULD SETUP THE DEVICE NODES
 *
 *		If using the device file system then set up the device
 *		nodes for this port. Ports above 32 use the extended minor
 *		number bits (in the upper 2 bytes of the device number).
 */

/*
		unit = strtport + portp->portnr;
		mindev = ((portp->brdnr & 0x7) << 20) | 
			(unit & 0x1f) | ((unit & 0x20) << 11);
		unit += (portp->brdnr * STL_PORTSPERBRD);

		portp->devfs_ttyE = devfs_add_devswf(&stl_cdevsw,
			mindev, DV_CHR,
			UID_ROOT, GID_WHEEL, 0600, "ttyE%d", unit);
		portp->devfs_ttyiE = devfs_add_devswf(&stl_cdevsw,
			(mindev | STL_CTRLINIT), DV_CHR,
			UID_ROOT, GID_WHEEL, 0600, "ttyiE%d", unit);
		portp->devfs_ttylE = devfs_add_devswf(&stl_cdevsw,
			(mindev | STL_CTRLLOCK), DV_CHR,
			UID_ROOT, GID_WHEEL, 0600, "ttylE%d", unit);
		portp->devfs_cue = devfs_add_devswf(&stl_cdevsw,
			(mindev | STL_CALLOUTDEV), DV_CHR,
			UID_UUCP, GID_DIALER, 0660, "cue%d", unit);
		portp->devfs_cuie = devfs_add_devswf(&stl_cdevsw,
			(mindev | STL_CTRLINIT | STL_CALLOUTDEV), DV_CHR,
			UID_UUCP, GID_DIALER, 0660, "cuie%d", unit);
		portp->devfs_cule = devfs_add_devswf(&stl_cdevsw,
			(mindev | STL_CTRLLOCK | STL_CALLOUTDEV), DV_CHR,
			UID_UUCP, GID_DIALER, 0660, "cule%d", unit);
*/

	}

	return(0);
}

/*****************************************************************************/

/*
 *	If we haven't scheduled a timeout then do it, some port needs high
 *	level processing.
 */

void stl_dotimeout()
{
#if DEBUG
	printf("stl_dotimeout()\n");
#endif

	if (stl_doingtimeout == 0) {
		timeout(stl_poll, 0, 1);
		stl_doingtimeout++;
	}
}

/*****************************************************************************/

/*
 *	Service "software" level processing. Too slow or painfull to be done
 *	at real hardware interrupt time. This way we might also be able to
 *	do some service on other waiting ports as well...
 */

void stl_poll(void *arg)
{
	return;
/*
	stlbrd_t	*brdp;
	stlport_t	*portp;
	struct tty	*tp;
	int		brdnr, portnr, rearm, x;

#if DEBUG
	printf("stl_poll()\n");
#endif

	stl_doingtimeout = 0;
	rearm = 0;

	x = spltty();
	for (brdnr = 0; (brdnr < stl_nrbrds); brdnr++) {
		if ((brdp = stl_brds[brdnr]) == (stlbrd_t *) NULL)
			continue;
		for (portnr = 0; (portnr < brdp->nrports); portnr++) {
			if ((portp = brdp->ports[portnr]) == (stlport_t *) NULL)
				continue;
			if ((portp->state & ASY_ACTIVE) == 0)
				continue;
			tp = &portp->tty;

			if (portp->state & ASY_RXDATA)
				stl_rxprocess(portp);
			if (portp->state & ASY_DCDCHANGE) {
				portp->state &= ~ASY_DCDCHANGE;
				portp->sigs = stl_getsignals(portp);
				(*linesw[tp->t_line].l_modem)(tp,
					(portp->sigs & TIOCM_CD));
			}
			if (portp->state & ASY_TXEMPTY) {
				if (stl_datastate(portp) == 0) {
					portp->state &= ~ASY_TXEMPTY;
					tp->t_state &= ~TS_BUSY;
					(*linesw[tp->t_line].l_start)(tp);
				}
			}
			if (portp->state & ASY_TXLOW) {
				portp->state &= ~ASY_TXLOW;
				(*linesw[tp->t_line].l_start)(tp);
			}

			if (portp->state & ASY_ACTIVE)
				rearm++;
		}
	}
	splx(x);

	if (rearm)
		stl_dotimeout();
*/
}

/*
 *	Start (or continue) the transfer of TX data on this port. If the
 *	port is not currently busy then load up the interrupt ring queue
 *	buffer and kick of the transmitter. If the port is running low on
 *	TX data then refill the ring queue. This routine is also used to
 *	activate input flow control.
 */

static void stl_start(struct tty *tp)
{
	stlport_t	*portp;
	unsigned int	len, stlen;
	char		*head, *tail;
	int		count, x;

	portp = (stlport_t *) tp;

#if DEBUG
	printf("stl_start(tp=%x): brdnr=%d portnr=%d\n", (int) tp, 
	    portp->brdnr, portp->portnr);
#endif

	x = spltty();

/*
 *	Check if the ports input has been blocked, and take appropriate action.
 *	Not very often do we really need to do anything, so make it quick.
 */
	if (tp->t_state & TS_TBLOCK) {
		if ((portp->state & ASY_RTSFLOWMODE) &&
		    ((portp->state & ASY_RTSFLOW) == 0))
			stl_flowcontrol(portp, 0, -1);
	} else {
		if (portp->state & ASY_RTSFLOW)
			stl_flowcontrol(portp, 1, -1);
	}

#if VFREEBSD == 205
/*
 *	Check if the output cooked clist buffers are near empty, wake up
 *	the line discipline to fill it up.
 */
	if (tp->t_outq.c_cc <= tp->t_lowat) {
		if (tp->t_state & TS_ASLEEP) {
			tp->t_state &= ~TS_ASLEEP;
			wakeup(&tp->t_outq);
		}
		selwakeup(&tp->t_wsel);
	}
#endif

	if (tp->t_state & (TS_TIMEOUT | TS_TTSTOP)) {
		splx(x);
		return;
	}

/*
 *	Copy data from the clists into the interrupt ring queue. This will
 *	require at most 2 copys... What we do is calculate how many chars
 *	can fit into the ring queue, and how many can fit in 1 copy. If after
 *	the first copy there is still more room then do the second copy. 
 *	The beauty of this type of ring queue is that we do not need to
 *	spl protect our-selves, since we only ever update the head pointer,
 *	and the interrupt routine only ever updates the tail pointer.
 */
	if (tp->t_outq.c_cc != 0) {
		head = portp->tx.head;
		tail = portp->tx.tail;
		if (head >= tail) {
			len = STL_TXBUFSIZE - (head - tail) - 1;
			stlen = portp->tx.endbuf - head;
		} else {
			len = tail - head - 1;
			stlen = len;
		}

		if (len > 0) {
			stlen = MIN(len, stlen);
			count = q_to_b(&tp->t_outq, head, stlen);
			len -= count;
			head += count;
			if (head >= portp->tx.endbuf) {
				head = portp->tx.buf;
				if (len > 0) {
					stlen = q_to_b(&tp->t_outq, head, len);
					head += stlen;
					count += stlen;
				}
			}
			portp->tx.head = head;
			if (count > 0)
				stl_startrxtx(portp, -1, 1);
		}

/*
 *		If we sent something, make sure we are called again.
 */
		tp->t_state |= TS_BUSY;
	}

#if VFREEBSD != 205
/*
 *	Do any writer wakeups.
 */
	ttwwakeup(tp);
#endif

	splx(x);
}

static int stl_param(struct tty *tp, struct termios *tiosp)
{
	stlport_t	*portp;

	portp = (stlport_t *) tp;
	if (portp == (stlport_t *) NULL)
		return(ENODEV);

	return(stl_setport(portp, tiosp));
}

/*
 *	Action the flow control as required. The hw and sw args inform the
 *	routine what flow control methods it should try.
 */

static void stl_flowcontrol(stlport_t *portp, int hw, int sw)
{
	unsigned char	*head, *tail;
	int		len, hwflow;

#if DEBUG
	printf("stl_flowcontrol(portp=%x,hw=%d,sw=%d)\n", (int) portp, hw, sw);
#endif

	hwflow = -1;

	if (portp->state & ASY_RTSFLOWMODE) {
		if (hw == 0) {
			if ((portp->state & ASY_RTSFLOW) == 0)
				hwflow = 0;
		} else if (hw > 0) {
			if (portp->state & ASY_RTSFLOW) {
				head = portp->rx.head;
				tail = portp->rx.tail;
				len = (head >= tail) ? (head - tail) :
					(STL_RXBUFSIZE - (tail - head));
				if (len < STL_RXBUFHIGH)
					hwflow = 1;
			}
		}
	}

/*
 *	We have worked out what to do, if anything. So now apply it to the
 *	UART port.
 */
	stl_sendflow(portp, hwflow, sw);
}

/*
 *	Initialize the port hardware. This involves enabling the transmitter
 *	and receiver, setting the port configuration, and setting the initial
 *	signal state.
 */

static int stl_rawopen(stlport_t *portp)
{
#if DEBUG
	printf("stl_rawopen(portp=%x): brdnr=%d panelnr=%d portnr=%d\n",
	    portp, portp->brdnr, portp->panelnr, portp->portnr);
#endif

	stl_setport(portp, &portp->tty.t_termios);
	portp->sigs = stl_getsignals(portp);
	stl_setsignals(portp, 1, 1);
	stl_enablerxtx(portp, 1, 1);
	stl_startrxtx(portp, 1, 0);
	return(0);
}

/*
 *	Shutdown the hardware of a port. Disable its transmitter and
 *	receiver, and maybe drop signals if appropriate.
 */

static int stl_rawclose(stlport_t *portp)
{
	struct tty	*tp;

#if DEBUG
	printf("stl_rawclose(portp=%x): brdnr=%d panelnr=%d portnr=%d\n",
	    portp, portp->brdnr, portp->panelnr, portp->portnr);
#endif

	tp = &portp->tty;
	stl_disableintrs(portp);
	stl_enablerxtx(portp, 0, 0);
	stl_flush(portp, (FWRITE | FREAD));
	if (tp->t_cflag & HUPCL) {
		stl_setsignals(portp, 0, 0);
		if (portp->dtrwait != 0) {
			portp->state |= ASY_DTRWAIT;
			timeout(stl_dtrwakeup, portp, portp->dtrwait);
		}
	}
	portp->callout = 0;
	portp->brklen = 0;
	portp->state &= ~(ASY_ACTIVE | ASY_RTSFLOW);
	wakeup(&portp->callout);
	wakeup(TSA_CARR_ON(tp));
	return(0);
}


static void stl_flush(stlport_t *portp, int flag)
{
	char	*head, *tail;
	int	len, x;

#if DEBUG
	printf("stl_flush(portp=%x,flag=%x)\n", (int) portp, flag);
#endif

	if (portp == (stlport_t *) NULL)
		return;

	x = spltty();

	if (flag & FWRITE) {
		stl_uartflush(portp, FWRITE);
		portp->tx.tail = portp->tx.head;
	}

/*
 *	The only thing to watch out for when flushing the read side is
 *	the RX status buffer. The interrupt code relys on the status
 *	bytes as being zeroed all the time (it does not bother setting
 *	a good char status to 0, it expects that it already will be).
 *	We also need to un-flow the RX channel if flow control was
 *	active.
 */
	if (flag & FREAD) {
		head = portp->rx.head;
		tail = portp->rx.tail;
		if (head != tail) {
			if (head >= tail) {
				len = head - tail;
			} else {
				len = portp->rx.endbuf - tail;
				bzero(portp->rxstatus.buf,
					(head - portp->rx.buf));
			}
			bzero((tail + STL_RXBUFSIZE), len);
			portp->rx.tail = head;
		}
		stl_uartflush(portp, FREAD);

		if ((portp->state & ASY_RTSFLOW) &&
		    ((portp->tty.t_state & TS_TBLOCK) == 0))
			stl_flowcontrol(portp, 1, -1);
	}

	splx(x);
}
/*
 *	Clear the DTR waiting flag, and wake up any sleepers waiting for
 *	DTR wait period to finish.
 */

static void stl_dtrwakeup(void *arg)
{
	stlport_t	*portp;

	portp = (stlport_t *) arg;
	portp->state &= ~ASY_DTRWAIT;
	wakeup(&portp->dtrwait);
}

/* stl_stop.  This is used as a function pointer in the tty struct for t_stop */
static void stl_stop(struct tty *tp, int rw)
{
	uprintf("stl_stop(tp=%x,rw=%x)\n", (int) tp, rw);

	stl_flush((stlport_t *) tp, rw);
}
