/*
 * Copyright (c) 2001 Murray Stokely
 * Copyright (c) 2001 Wind River Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <dev/stl/scd1400.h>
#include <dev/stl/sc26198.h>

#define	MIN(a,b) (((a)<(b))?(a):(b))

/* XXX FIXME TEMPORARY SHORTCUT */
#if __FreeBSD_version >= 500023
#define SUSER suser_td(td)
#else
#define SUSER suser(p)
#endif

#define	STL_VENDOR_ID      0x124d
#define STL_EASYIO_DEVID   0x0003

/*
 *	Define important driver limitations.
 */
#define	STL_MAXBRDS		8
#define	STL_MAXPANELS		4
#define	STL_MAXBANKS		8
#define	STL_PORTSPERPANEL	16
#define	STL_PORTSPERBRD		64
#define STL_DEVSPERPORT         6

#define MINORTOPORT(unit, minor) \
  ((unit > 0) ? \
     (((minor) % ((unit) * (STL_PORTSPERBRD) * (STL_DEVSPERPORT))) / (STL_DEVSPERPORT)) : \
     (minor > 0 ? \
       (minor / STL_DEVSPERPORT) : \
       (minor)))
#define MINORTODEV(unit, minor) \
  ((unit > 0) ? \
    (((MINORTOPORT(unit, minor)) > 0) ? \
      (((minor) % ((unit) * (STL_PORTSPERBRD) * (STL_DEVSPERPORT))) % (MINORTOPORT(unit, minor) * STL_DEVSPERPORT)) :\
      ((minor) % ((unit) * (STL_PORTSPERBRD) * (STL_DEVSPERPORT)))) : \
    (minor > 0 ? \
      (((MINORTOPORT(unit, minor)) > 0) ? \
        ((minor) % (MINORTOPORT(unit, minor) * STL_DEVSPERPORT)) : \
        (minor)) : \
      (minor)))
#define MINORTOUNIT(minor) \
  ((minor) / ((STL_PORTSPERBRD) * (STL_DEVSPERPORT)))
#define STL_MINOR(unit, port, dev) \
   (((unit) * (STL_PORTSPERBRD) * (STL_DEVSPERPORT)) + ((port) * (STL_DEVSPERPORT)) + dev)

/*
 *	Define some macros to make calling these functions nice and clean.
 */
#define	stl_panelinit		(* ((uart_t *) panelp->uartp)->panelinit)
#define	stl_portinit		(* ((uart_t *) portp->uartp)->portinit)
#define	stl_setport		(* ((uart_t *) portp->uartp)->setport)
#define	stl_getsignals		(* ((uart_t *) portp->uartp)->getsignals)
#define	stl_setsignals		(* ((uart_t *) portp->uartp)->setsignals)
#define	stl_enablerxtx		(* ((uart_t *) portp->uartp)->enablerxtx)
#define	stl_startrxtx		(* ((uart_t *) portp->uartp)->startrxtx)
#define	stl_disableintrs	(* ((uart_t *) portp->uartp)->disableintrs)
#define	stl_sendbreak		(* ((uart_t *) portp->uartp)->sendbreak)
#define	stl_sendflow		(* ((uart_t *) portp->uartp)->sendflow)
#define	stl_uartflush		(* ((uart_t *) portp->uartp)->flush)
#define	stl_datastate		(* ((uart_t *) portp->uartp)->datastate)




/*****************************************************************************/

/*
 *	Hardware ID bits for the EasyIO and ECH boards. These defines apply
 *	to the directly accessable io ports of these boards (not the cd1400
 *	or 26198 uarts - they are in scd1400.h and sc26198.h).
 */
#define	EIO_8PORTRS	0x04
#define	EIO_4PORTRS	0x05
#define	EIO_8PORTDI	0x00
#define	EIO_8PORTM	0x06
#define	EIO_MK3		0x03
#define	EIO_IDBITMASK	0x07

#define	EIO_BRDMASK	0xf0
#define	ID_BRD4		0x10
#define	ID_BRD8		0x20
#define	ID_BRD16	0x30

#define	EIO_INTRPEND	0x08
#define	EIO_INTEDGE	0x00
#define	EIO_INTLEVEL	0x08

#define	ECH_ID		0xa0
#define	ECH_IDBITMASK	0xe0
#define	ECH_BRDENABLE	0x08
#define	ECH_BRDDISABLE	0x00
#define	ECH_INTENABLE	0x01
#define	ECH_INTDISABLE	0x00
#define	ECH_INTLEVEL	0x02
#define	ECH_INTEDGE	0x00
#define	ECH_INTRPEND	0x01
#define	ECH_BRDRESET	0x01

#define	ECHMC_INTENABLE	0x01
#define	ECHMC_BRDRESET	0x02

#define	ECH_PNLSTATUS	2
#define	ECH_PNL16PORT	0x20
#define	ECH_PNLIDMASK	0x07
#define	ECH_PNLXPID	0x40
#define	ECH_PNLINTRPEND	0x80
#define	ECH_ADDR2MASK	0x1e0

#define	EIO_CLK		25000000
#define	EIO_CLK8M	20000000
#define	ECH_CLK		EIO_CLK

/*
 *	Define different board types. At the moment I have only declared
 *	those boards that this driver supports. But I will use the standard
 *	"assigned" board numbers. In the future this driver will support
 *	some of the other Stallion boards. Currently supported boards are
 *	abbreviated as EIO = EasyIO and ECH = EasyConnection 8/32.
 */
#define	BRD_EASYIO	20
#define	BRD_ECH		21
#define	BRD_ECHMC	22
#define	BRD_ECHPCI	26
#define	BRD_ECH64PCI	27
#define	BRD_EASYIOPCI	28


/*
 *	Define the important minor number break down bits. These have been
 *	chosen to be "compatible" with the standard sio driver minor numbers.
 *	Extra high bits are used to distinguish between boards.
 */
#define	STL_CALLOUTDEV		0x80
#define	STL_CTRLLOCK		0x40
#define	STL_CTRLINIT		0x20
#define	STL_CTRLDEV		(STL_CTRLLOCK | STL_CTRLINIT)

#define	STL_MEMDEV		0x07000000

#define	STL_DEFSPEED		9600
#define	STL_DEFCFLAG		(CS8 | CREAD | HUPCL)

/*
 *	I haven't really decided (or measured) what buffer sizes give
 *	a good balance between performance and memory usage. These seem
 *	to work pretty well...
 */
#define	STL_RXBUFSIZE		2048
#define	STL_TXBUFSIZE		2048

#define	STL_TXBUFLOW		(STL_TXBUFSIZE / 4)
#define	STL_RXBUFHIGH		(3 * STL_RXBUFSIZE / 4)

/*
 *	Per board state flags. Used with the state field of the board struct.
 *	Not really much here yet!
 */
#define	BRD_FOUND	0x1

/*
 *	Define the port structure state flags. These set of flags are
 *	modified at interrupt time - so setting and reseting them needs
 *	to be atomic.
 */
#define	ASY_TXLOW	0x1
#define	ASY_RXDATA	0x2
#define	ASY_DCDCHANGE	0x4
#define	ASY_DTRWAIT	0x8
#define	ASY_RTSFLOW	0x10
#define	ASY_RTSFLOWMODE	0x20
#define	ASY_CTSFLOWMODE	0x40
#define	ASY_TXFLOWED	0x80
#define	ASY_TXBUSY	0x100
#define	ASY_TXEMPTY	0x200

#define	ASY_ACTIVE	(ASY_TXEMPTY | ASY_TXLOW | ASY_RXDATA | ASY_DCDCHANGE)


/*
 *	Define a ring queue structure for each port. This will hold the
 *	TX data waiting to be output. Characters are fed into this buffer
 *	from the line discipline (or even direct from user space!) and
 *	then fed into the UARTs during interrupts. Will use a clasic ring
 *	queue here for this. The good thing about this type of ring queue
 *	is that the head and tail pointers can be updated without interrupt
 *	protection - since "write" code only needs to change the head, and
 *	interrupt code only needs to change the tail.
 */
typedef struct {
	char	*buf;
	char	*endbuf;
	char	*head;
	char	*tail;
} stlrq_t;

/*
 *	Port, panel and board structures to hold status info about each.
 *	The board structure contains pointers to structures for each panel
 *	connected to it, and in turn each panel structure contains pointers
 *	for each port structure for each port on that panel. Note that
 *	the port structure also contains the board and panel number that it
 *	is associated with, this makes it (fairly) easy to get back to the
 *	board/panel info for a port. Also note that the tty struct is at
 *	the top of the structure, this is important, since the code uses
 *	this fact to get the port struct pointer from the tty struct
 *	pointer!
 */
typedef struct stlport {
	struct tty	tty;
	int		portnr;
	int		panelnr;
        struct stl_softc *sc;
	int		ioaddr;
	int		uartaddr;
	int		pagenr;
	int		callout;
	int		brklen;
	int		dtrwait;
	int		dotimestamp;
	int		waitopens;
	int		hotchar;
	void		*uartp;
	unsigned int	state;
	unsigned int	hwid;
	unsigned int	sigs;
	unsigned int	rxignoremsk;
	unsigned int	rxmarkmsk;
	unsigned int	crenable;
	unsigned int	imr;
	unsigned long	clk;
	struct termios	initintios;
	struct termios	initouttios;
	struct termios	lockintios;
	struct termios	lockouttios;
	struct timeval	timestamp;
	comstats_t	stats;
	stlrq_t		tx;
	stlrq_t		rx;
	stlrq_t		rxstatus;
} stlport_t;

typedef struct stlpanel {
        int		panelnr;  /* Panel # */
        struct stl_softc *sc;
	int		pagenr;
	int		nrports;
	int		iobase;
	unsigned int	hwid;
	unsigned int	ackmask;
	void		(*isr)(struct stlpanel *panelp, unsigned int iobase);
	void		*uartp;
	stlport_t	*ports[STL_PORTSPERPANEL];
} stlpanel_t;

typedef struct stlbrd {
	int		brdnr;
	int		unitid;
	int		nrbnks;
	unsigned int	iostatus;
	void		(*isr)(struct stlbrd *brdp);
	unsigned int	bnkpageaddr[STL_MAXBANKS];
	unsigned int	bnkstataddr[STL_MAXBANKS];
	stlpanel_t	*bnk2panel[STL_MAXBANKS];
} stlbrd_t;

struct stl_softc {
#if __FreeBSD_version > 500013
    struct mtx lock;
#endif
    device_t dev;

    struct resource *bar0res;	/* BAR0 */
    int bar0id;
    bus_space_tag_t bar0_bt;
    bus_space_handle_t bar0_bh;

    struct resource *bar1res;	/* BAR1 */
    int bar1id;
    bus_space_tag_t bar1_bt;
    bus_space_handle_t bar1_bh;

    struct resource *irqres;    /* IRQ */
    int irqid;
    bus_space_tag_t irq_bt;
    bus_space_handle_t irq_bh;
    void *handler;

    struct stl_device *device_info;

    dev_t iomemdev;
    dev_t ttyE[STL_PORTSPERBRD];
    dev_t ttyiE[STL_PORTSPERBRD];
    dev_t ttylE[STL_PORTSPERBRD];
    dev_t cue[STL_PORTSPERBRD];
    dev_t cuie[STL_PORTSPERBRD];
    dev_t cule[STL_PORTSPERBRD];

    stlpanel_t	        *panels[STL_MAXPANELS];
    stlport_t	        *ports[STL_PORTSPERBRD];
    int		        nrpanels;
    int		        nrports;
    unsigned int	ioctrl;
    unsigned int	hwid;
    int		        state;
    unsigned long	clk;
    int		        brdtype;
    unsigned int	ioctrlval;

    int                 unit; /* Unit number of this device */
};

/*****************************************************************************/

/*
 *	Generic UART support structure.
 */
typedef struct uart {
    int	(*panelinit)(struct stl_softc *sc, stlpanel_t *panelp);
	void	(*portinit)(struct stl_softc *sc, stlpanel_t *panelp, stlport_t *portp);
	int	(*setport)(stlport_t *portp, struct termios *tiosp);
	int	(*getsignals)(stlport_t *portp);
	void	(*setsignals)(stlport_t *portp, int dtr, int rts);
	void	(*enablerxtx)(stlport_t *portp, int rx, int tx);
	void	(*startrxtx)(stlport_t *portp, int rx, int tx);
	void	(*disableintrs)(stlport_t *portp);
	void	(*sendbreak)(stlport_t *portp, long len);
	void	(*sendflow)(stlport_t *portp, int hw, int sw);
	void	(*flush)(stlport_t *portp, int flag);
	int	(*datastate)(stlport_t *portp);
	void	(*intr)(stlpanel_t *panelp, unsigned int iobase);
} uart_t;


/* Board descriptions. */
struct stl_device {
    uint16_t device_id;
    char description[80];
    int ports;
};

/*
 *	Set up enable and disable macros for the ECH boards. They require
 *	the secondary io address space to be activated and deactivated.
 *	This way all ECH boards can share their secondary io region.
 *	If this is an ECH-PCI board then also need to set the page pointer
 *	to point to the correct page.
 */
#define	BRDENABLE(sc, pagenr)						\
	if (sc->brdtype == BRD_ECH)			\
		BAR0_WRITE1(sc, 1, (sc->ioctrlval | ECH_BRDENABLE));\
	else if (sc->brdtype == BRD_ECHPCI)		\
		BAR0_WRITE1(sc, 1, (pagenr));

/* This assumes ioctrl = bar0 + 1 */
#define	BRDDISABLE(sc)						\
	if (sc->brdtype == BRD_ECH)			\
		BAR0_WRITE1(sc, 1, (sc->ioctrlval | ECH_BRDDISABLE));

/* Function prototypes */

extern devclass_t stl_devclass;

int stl_makedevs(struct stl_softc *sc);
int stl_brdinit(struct stl_softc *sc);
/*
static int stl_probe(device_t dev);
static int stl_attach(device_t dev);
static int stl_detach(device_t dev);
static int stl_shutdown(device_t dev);
*/
void stl_handler(void *xsc);

#define BAR0_WRITE1(sc, address, value) \
        bus_space_write_1(sc->bar0_bt, sc->bar0_bh, address, value)
#define BAR0_WRITE2(sc, address, value) \
        bus_space_write_2(sc->bar0_bt, sc->bar0_bh, address, value)
#define BAR1_WRITE1(sc, address, value) \
        bus_space_write_1(sc->bar1_bt, sc->bar1_bh, address, value)
#define BAR1_WRITE2(sc, address, value) \
        bus_space_write_2(sc->bar1_bt, sc->bar1_bh, address, value)
#define BAR0_READ1(sc, address) \
        bus_space_read_1(sc->bar0_bt, sc->bar0_bh, address)
#define BAR0_READ2(sc, address) \
        bus_space_read_2(sc->bar0_bt, sc->bar0_bh, address)
#define BAR1_READ1(sc, address) \
        bus_space_read_1(sc->bar1_bt, sc->bar1_bh, address)
#define BAR1_READ2(sc, address) \
        bus_space_read_2(sc->bar1_bt, sc->bar1_bh, address)

/*
 *	SC26198 uart specific handling functions.
 */
void	stl_sc26198setreg(stlport_t *portp, int regnr, int value);
int	stl_sc26198getreg(stlport_t *portp, int regnr);
int	stl_sc26198updatereg(stlport_t *portp, int regnr, int value);
int	stl_sc26198getglobreg(stlport_t *portp, int regnr);
int     stl_sc26198panelinit(struct stl_softc *sc, stlpanel_t *panelp);
void stl_sc26198portinit(struct stl_softc *sc, stlpanel_t *panelp, stlport_t *portp);
int	stl_sc26198setport(stlport_t *portp, struct termios *tiosp);
int	stl_sc26198getsignals(stlport_t *portp);
void	stl_sc26198setsignals(stlport_t *portp, int dtr, int rts);
void	stl_sc26198enablerxtx(stlport_t *portp, int rx, int tx);
void	stl_sc26198startrxtx(stlport_t *portp, int rx, int tx);
void	stl_sc26198disableintrs(stlport_t *portp);
void	stl_sc26198sendbreak(stlport_t *portp, long len);
void	stl_sc26198sendflow(stlport_t *portp, int hw, int sw);
int	stl_sc26198datastate(stlport_t *portp);
void	stl_sc26198flush(stlport_t *portp, int flag);
void	stl_sc26198txunflow(stlport_t *portp);
void	stl_sc26198wait(stlport_t *portp);
void	stl_sc26198intr(stlpanel_t *panelp, unsigned int iobase);
void	stl_sc26198txisr(stlport_t *port);
void	stl_sc26198rxisr(stlport_t *port, unsigned int iack);
void	stl_sc26198rxgoodchars(stlport_t *portp);
void	stl_sc26198rxbadchars(stlport_t *portp);
void	stl_sc26198otherisr(stlport_t *port, unsigned int iack);

/*
 *	Define the offsets within the register bank of a cd1400 based panel.
 *	These io address offsets are common to the EasyIO board as well.
 */
#define	EREG_ADDR	0
#define	EREG_DATA	4
#define	EREG_RXACK	5
#define	EREG_TXACK	6
#define	EREG_MDACK	7

#define	EREG_BANKSIZE	8

#define	CD1400_CLK	25000000
#define	CD1400_CLK8M	20000000



/*****************************************************************************/

/*
 *	SC26198 UART specific data initization.
 */
static uart_t stl_sc26198uart = {
	stl_sc26198panelinit,
	stl_sc26198portinit,
	stl_sc26198setport,
	stl_sc26198getsignals,
	stl_sc26198setsignals,
	stl_sc26198enablerxtx,
	stl_sc26198startrxtx,
	stl_sc26198disableintrs,
	stl_sc26198sendbreak,
	stl_sc26198sendflow,
	stl_sc26198flush,
	stl_sc26198datastate,
	stl_sc26198intr
};

/*
 *	Define the offsets within the register bank of a sc26198 based panel.
 */
#define	XP_DATA		0
#define	XP_ADDR		1
#define	XP_MODID	2
#define	XP_STATUS	2
#define	XP_IACK		3

#define	XP_BANKSIZE	4

/*
 *	Define the sc26198 baud rate table. Offsets within the table
 *	represent the actual baud rate selector of sc26198 registers.
 */
static unsigned int	sc26198_baudtable[] = {
	50, 75, 150, 200, 300, 450, 600, 900, 1200, 1800, 2400, 3600,
	4800, 7200, 9600, 14400, 19200, 28800, 38400, 57600, 115200,
	230400, 460800
};

#define	SC26198_NRBAUDS		(sizeof(sc26198_baudtable) / sizeof(unsigned int))

/*
 *	Define the maximum baud rate of the sc26198 devices.
 */
#define	SC26198_MAXBAUD	460800

void stl_ttyoptim(stlport_t *portp, struct termios *tiosp);
void stl_dotimeout(void);
void stl_poll(void *arg);
